/*
 *			dictlink.c
 * $Id: dictlink.c,v 1.14 2004/01/22 10:17:38 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "debug.h"
#include "memory.h"
#include "wordhist.h"
#include "property.h"
#include "dictbook.h"
#include "booklist.h"
#include "subr.h"
#include "dictlink.h"
#include "dlentlist.c"


/* static variables in dictlink.c */
static int listview = 1;	/* list view option flag */
static int alwaysopen = 1;	/* always-open flag */
static dictlink_error_handler_t errhndlr = NULL; /* err handler */
static void (*viewfree)(void *) = NULL;
void dictlink_set_viewfree(void (*v)(void *)){ viewfree = v; }


/* static functions */
static void 	dl_free_node(struct dictnode *xdp);
static void 	free_dlistent(struct dlistent *);
/* static void    	remove_book_on_error(booklist, struct dictbook *, 
				     struct dictnode *);*/
static strlist 	lookup_word(booklist, struct dictbook *, 
			    const char *, struct dictnode *);
static strlist 	get_entry(booklist, struct dictbook *, 
			  const char *, struct dictnode *);

static dlentlist lookup_word_list(booklist, const char *, 
				  struct dictnode *);
static struct dictnode *construct_children(dlentlist, struct dictnode *);

#define open_book(b, x)		do { 			 \
		if(alwaysopen == 0 && dict_open(b) == 0) \
			return x;			 \
	} while(0)

#define close_book(b)		do { 			\
		if(alwaysopen == 0)			\
			dict_close(b);			\
	} while(0)
		

/*----------------------------------------------------------------*/
	/* definitions of external functions */
void 
dictlink_init(booklist books, dictlink_error_handler_t h)
{ 
	if(DEBUG_IS_ENABLED(DEBUG_DICTLINK)){
		booklist_iterator itr = iterator_first(books);
		printf("(dictlink): dictlink_init()\n");
		for(; !iterator_is_end(itr); itr = iterator_next(itr)){
			struct dictbook *b = iterator_data(itr);
			printf(" - %s %s %x\n", b->db_glid, b->db_name, b->db_flags); 
		}
	}
	/* load word lookup history database */
	whlookup_load(); 

	/* get properties */
	listview = prty_integer_val(property_get("ListView"));
	alwaysopen = prty_integer_val(property_get("AlwaysOpen"));

	/* open first. they will be closed at dictlink_fin() */
	if(alwaysopen){
		booklist_iterator itr = iterator_first(books);
		for(; !iterator_is_end(itr); itr = iterator_next(itr))
			dict_open(iterator_data(itr));
	}

	errhndlr = h;
}

void 
dictlink_fin(booklist books)
{
	if(DEBUG_IS_ENABLED(DEBUG_DICTLINK))
		printf("(dictlink): dictlink_fin()\n");
	/* always open options is enabled, all books must be closed here */
	if(alwaysopen){
		booklist_iterator itr = iterator_first(books);
		for(; !iterator_is_end(itr); itr = iterator_next(itr))
			dict_close(iterator_data(itr));
	}

	whlookup_fin();
}

int
dl_has_up(struct dictnode *current)
{
	/* on list-view, all entries are able to be shown */
	if(listview)
		return (current->dn_parent != NULL);

	/* on non-list-view */
	if(!current->dn_parent)
		return 0;	/* there is no parent */

	/* list-node cannot be shown  */
	if(current->dn_parent->dn_type != DNTYPE_LIST)
		return 1;	/* TRANS node can be shown */

	/* ask recursively */
	return dl_has_up(current->dn_parent);
}

int
dl_has_next(struct dictnode *current)
{
	return (current->dn_type != DNTYPE_LIST && current->dn_next != NULL);
}
int
dl_has_prev(struct dictnode *current)
{
	return (current->dn_type != DNTYPE_LIST && current->dn_prev != NULL);
}

struct dictnode *
dl_up(booklist blist, struct dictnode *current)
{
	struct dictnode *pp = current->dn_parent;

	/* there is no node which could be parent */
	if(!dl_has_up(current))
		return NULL;	

	/* if current node type is DNTYPE_LIST, just free and return parent */

	/* on list-view, all entries are able to be shown */
	if(listview){
		/* current node type is list */
		if(current->dn_type == DNTYPE_LIST)
			dl_free_node(current);

		/* 
		 *  current node is translation-node and 
		 * parent node type is translation-node. 
		 */
		else if(current->dn_parent->dn_type == DNTYPE_TRANS)
			 dl_free_node(current);

		/*  only if current node type is translation and type of 
		 * parent is list, the child trans-nodes will be freed by
		 * parent list-node. 
		 */
		return pp;
	}
			
	/* not list-view */

	/* type of parent is not list */
	if(current->dn_parent->dn_type == DNTYPE_TRANS){
		dl_free_node(current);
		return pp;
	}

	/* if parent type is DNTYPE_LIST on non-list-view, check grandparent */
	if(pp->dn_parent){
		struct dictnode *ret = pp->dn_parent;

		/*   free only list-node of parent. list-node will free
		 *  child nodes.
		 */
		dl_free_node(pp);  
		return ret;
	}

	/* should not reach here */
	return NULL;
}

struct dictnode *
dl_next(booklist blist, struct dictnode *current)
{
	struct dictnode *np = current->dn_next;

	if(!np)
		return NULL;

	/* contents of translation is not read yet */
	if(!np->dn_content){
		np->dn_content = get_entry(blist, np->dn_book, 
					   np->dn_key, current);
		if(np->dn_content == NULL)
			return NULL; /* error occurs */
	}
	return np;
}

struct dictnode *
dl_prev(booklist blist, struct dictnode *current)
{
	struct dictnode *np = current->dn_prev;

	if(!np)
		return NULL;

	/* contents of translation is not read yet */
	if(!np->dn_content){
		np->dn_content = get_entry(blist, np->dn_book, 
					   np->dn_key, current);
		if(np->dn_content == NULL)
			return NULL;
	}
	return np;
}

void
dl_quit(booklist books, struct dictnode *cur)
{
	booklist_iterator bitr;

	while(cur){
		struct dictnode *prevp = cur;
		cur = cur->dn_parent;

		/* prevp is trans-node */
		if(prevp->dn_type == DNTYPE_TRANS){
			/* if parent of trans-node is not list */
			if(prevp->dn_parent->dn_type != DNTYPE_LIST)
				dl_free_node(prevp); /* free prevp */
		}
		else if(prevp->dn_type == DNTYPE_LIST)
			dl_free_node(prevp);
	}

	/* save word lookup history */
	whlookup_save();

	/* save cache and free memory space */
	bitr = iterator_first(books);
	for(; !iterator_is_end(bitr); bitr = iterator_next(bitr))
		dict_reclaim(iterator_data(bitr));
}

unsigned int 
dl_list_count(struct dictnode *current)
{
	if(current->dn_type != DNTYPE_LIST)
		abort(); /* must be LIST node */

	return iterator_count((dlentlist)current->dn_entdata);
}

struct dlistent *
dl_list_at(struct dictnode *current, int idx)
{
	dlentlist lp;
	dlentlist_iterator itr;

	if(current->dn_type != DNTYPE_LIST)
		abort(); /* must be LIST node */

	lp = (dlentlist)current->dn_entdata;
	if(idx < 0 || idx >= iterator_count(lp))
		return NULL;

	itr = iterator_at(lp, idx);
	return iterator_data(itr);
}

struct dictnode *
dl_entry(booklist books, const char *entry, struct dictnode *current)
{
	struct dictnode *ret;
	strlist result;

	/* current node is list node */
	if(current->dn_type == DNTYPE_LIST){
		ret = current->dn_entry;
		for(; ret ; ret = ret->dn_next)
			if(!strcmp(ret->dn_key, entry))
				break;
		if(!ret)
			return NULL; /* no such chlidren */

		if(ret->dn_content)
			return ret;

		result = get_entry(books, ret->dn_book, ret->dn_key, current);

		/* error occured */
		if(result == NULL)
			return NULL;

		ret->dn_content = result;
		return ret;
	}

	/* current node is trans node */

	result = get_entry(books, current->dn_book, entry, current);

	/* error occured */
	if(!result)
		return NULL;

	ret = MALLOC("dictlink/trans", sizeof(struct dictnode));
	ret->dn_type = DNTYPE_TRANS;
	ret->dn_key = (const char *)xp_strdup(entry);
	ret->dn_content = result;
	ret->dn_book = current->dn_book;
	ret->dn_parent = current;
	ret->dn_whist = current->dn_whist;
	ret->dn_prev = ret->dn_next = NULL;
	ret->dn_number = current->dn_number;
	ret->dn_view = NULL;
	return ret;
}

struct dictnode *
dl_lookup(booklist books, const char *word, struct dictnode *current)
{
	struct dictnode 	*ret = NULL;
	dlentlist		dlist;
	struct whentry 		*wp;

	if(word == NULL || *word == '\0')
		return NULL;

	dlist = lookup_word_list(books, word, current);

	/* error or no such word */
	if(!dlist)
		return NULL;

	/* lookup counter */
	wp = whlookup_word(word);

	ret = MALLOC("dictnode/list", sizeof(struct dictnode));
	ret->dn_type = DNTYPE_LIST;
	ret->dn_whist = wp;
	ret->dn_key = wp->whe_word;
	ret->dn_parent = current;
	ret->dn_number = iterator_count(dlist);
	ret->dn_entdata = dlist;
	ret->dn_entry = construct_children(dlist, ret);
	ret->dn_view = NULL;

	/* disable list view */
	if(!listview){
		/* return first node of children */
		ret = ret->dn_entry;
		ret->dn_content = get_entry(books, ret->dn_book, 
						ret->dn_key, current);
	}

	if(DEBUG_IS_ENABLED(DEBUG_DICTLINK)){
		printf("(dictlink): dl_lookup(%s): return key:%s\n",
			word, ret->dn_key);
	}
	return ret;
}

/*----------------------------------------------------------------*/
/* definitions of external functions */

static void free_dlistent(struct dlistent *p)
{ 
	xp_strfree(p->dl_key);
	xp_strfree(p->dl_title);
	FREE("dlinstent", p); 
}

static void 	
dl_free_node(struct dictnode *xdp)
{
	if(xdp->dn_type == DNTYPE_LIST){
		/* free list of dlistent */
		aggregator_delete_by((dlentlist)xdp->dn_entdata, 
				     free_dlistent);
	
		/* free children */
		dl_free_node(xdp->dn_entry);

		/* free view */
		if(xdp->dn_view && viewfree)
			viewfree(xdp->dn_view);

		FREE("dictnode/list", xdp);
	}

	else if(xdp->dn_type == DNTYPE_TRANS){
		struct dictnode *next = NULL, *prev = NULL;

		/* remove out of the double-linked list */
		if(xdp->dn_next){
			xdp->dn_next->dn_prev = NULL;
			next = xdp->dn_next;
		}
		if(xdp->dn_prev){
			xdp->dn_prev->dn_next = NULL;
			prev = xdp->dn_prev;
		}

		/* free content */
		if(xdp->dn_content != NULL) 
			strlist_delete(xdp->dn_content);

		/* trans-node has duplicated string */
		xp_strfree(xdp->dn_key);

		/* free view */
		if(xdp->dn_view && viewfree)
			viewfree(xdp->dn_view);
		
		FREE("dictnode/trans", xdp);

		/* free siblings */
		if(next)
			dl_free_node(next);
		if(prev)
			dl_free_node(prev);
	}
}


#if 0	/* nouse */
/* 
 * remove_book_on_error():
 *    this function removes the book indicated by 2nd argument from 
 *   booklist and all dictnodes.
 */
static void    	
remove_book_on_error(booklist books, 
		     struct dictbook *book, 
		     struct dictnode *cur)
{
	booklist_iterator itr = iterator_find(books, book);

	if(iterator_is_end(itr))
		return;	/* this book is already removed */
	
	/* remove out of booklist */
	iterator_erase(itr);

	/* clear all pointer to this book */
	while(cur){
		struct dictnode *dp = cur->dn_parent;
		struct dictnode *pp = cur->dn_parent;
		dlentlist_iterator ditr;

		/* trans-node which isn't child of list-node */
		if(cur->dn_type == DNTYPE_TRANS){
			if(pp->dn_type != DNTYPE_LIST && 
			   cur->dn_book == book)
				cur->dn_book = NULL;

			cur = pp;
			continue;
		}

		/* List-node */

		/* check each entry on the list */
		ditr = iterator_first((dlentlist)current->dn_entdata);
		for(; !iterator_is_end(ditr); ditr = iterator_next(ditr))
			if(iterator_data(ditr)->dl_book == book)
				iterator_data(ditr)->dl_book = NULL;

		/* check children */
		for(dp = cur->dn_entry; dp ; dp = dp->dp_next)
			if(dp->dn_book == book)
				dp->dn_book = NULL;
		cur = pp;
	}

	/* close and free this book */
	dict_close(book);
	dict_free(book);

	/* no more books */
	if(iterator_count(books) == 0){
		if(errhndlr)
			errhndlr("xpopdic", "No available books", "Error");
		else
			fprintf(stderr, "xpopdic: No available books\n");
	}
}
#endif 


/* lookup word on a book */
static strlist 
lookup_word(booklist books, struct dictbook *book, 
	    const char *word, struct dictnode *dp)
{
	strlist sltp = NULL;

	open_book(book, NULL);
	sltp = dict_lookup(book, word); /* lookup */
	close_book(book);

	if(sltp != NULL)
		return sltp; /* success */

	if(dict_error(book)){
		char buf[256];
		snprintf(buf, 256, "lookup(%s)", word);

		if(errhndlr)
			errhndlr(book->db_glid, dict_strerr(book), buf);
		else
			fprintf(stderr, "xpopdic: dict_lookup(%s): %s\n", 
					book->db_glid, dict_strerr(book));
	}
	return NULL;
}

/* get translation entry on a book */
static strlist
get_entry(booklist books, struct dictbook *book, 
	  const char *entry, struct dictnode *cur)
{
	strlist sltp = NULL;

	open_book(book, NULL);
	sltp = dict_getentry(book, entry); /* get entry */
	close_book(book);

	if(sltp != NULL)
		return sltp; /* success */

	if(dict_error(book)){
		char buf[256];
		snprintf(buf, 256, "getentry(%s)", entry);

		if(errhndlr)
			errhndlr(book->db_glid, dict_strerr(book), buf);
		else
			fprintf(stderr, "xpopdic: dict_getentry(%s): %s\n", 
				book->db_glid, dict_strerr(book));
	}
	return NULL;
}

static dlentlist 
lookup_word_list(booklist books, const char *word, struct dictnode *current)
{
	booklist_iterator 	itr, next;
	dlentlist		ret = NULL;

	/* search the word for each dictionary */
	itr = iterator_first(books);
	for(; !iterator_is_end(itr); itr = next){
		struct dlistent *dep;
		struct dictbook *book;
		strlist		result;
		strlist_iterator sitr;

		/* 
		 *  When some error is occured, booklist will be changed. 
		 * To avoid crash, get next iterator before.
		 */
		next = iterator_next(itr);

		book = iterator_data(itr);
		result = lookup_word(books, book, word, current);

		if(!result)
			continue;

		/* dictionary entry list */
		if(ret == NULL)
			ret = dlentlist_new(4, 4, xpdic_malloc, xpdic_free);
	
		sitr = iterator_first(result);
		for(; !iterator_is_end(sitr); sitr = iterator_next(sitr)){
			const char *title = iterator_data(sitr);
			const char *entryid;
			sitr = iterator_next(sitr);
			entryid = iterator_data(sitr);

			dep = MALLOC("dlistent", sizeof(struct dlistent));
			dep->dl_title = title;
			dep->dl_key = entryid;
			dep->dl_book = book;
			iterator_append(ret, dep);
		}

		/* free only list. not each string */
		aggregator_delete(result);
	}
	return ret;
}

static struct dictnode *
construct_children(dlentlist dlist, struct dictnode *parent)
{
	int n = 0;
	struct dictnode *ret = NULL;
	struct dictnode *last = NULL;
	dlentlist_iterator  itr;

	itr = iterator_first(dlist);
	for(; !iterator_is_end(itr); itr = iterator_next(itr)){
		struct dictnode *p;
		struct dlistent *dep = iterator_data(itr);

		p = MALLOC("dictnode/trans", sizeof(struct dictnode));
		p->dn_type = DNTYPE_TRANS;
		p->dn_key = (const char *)xp_strdup(dep->dl_key);
		p->dn_book = dep->dl_book;
		p->dn_content = NULL;
		p->dn_parent = parent;
		p->dn_whist = parent->dn_whist;
		p->dn_prev = p->dn_next = NULL;
		p->dn_number = n++;
		p->dn_prev = p->dn_next = NULL;
		p->dn_view = NULL;

		/* add this entry into list */
		if(ret == NULL)
			ret = last = p;
		else {
			last->dn_next = p;
			p->dn_prev = last;
			last = p;
		}
	}

	return ret;
}

static void
dictnode_dump(struct dictnode *d, int n)
{
	int 			i;
	dlentlist_iterator  	itr;
	struct dlistent 	*dep;
	char 			idt[256];
	idt[0] = '\0';

	for(i = 0; i < n ; i++)
		strcat(idt, "  ");
	strcat(idt, "+-->");

	if(d == NULL)
		return;

	if(d->dn_type == DNTYPE_LIST){
		printf("%sList <%s> {\n", idt, d->dn_key);

		itr = iterator_first((dlentlist)d->dn_entdata); 
		for(; !iterator_is_end(itr); itr = iterator_next(itr)){
			dep = iterator_data(itr);
			printf("%s - %s@%s\n", idt, dep->dl_key, 
				dep->dl_book->db_glid);
		}
		printf("%s}\n", idt, d->dn_key);
	}
	else {
		printf("%sTrans <%s>@%s\n", idt, d->dn_key, 
				d->dn_book->db_glid);
	}
	dictnode_dump(d->dn_parent, n + 1);
}

static void test_handler(const char *dictname, const char *err, 
			 const char *cause)
{
	fprintf(stderr, "dictlink: %s: \"%s\" because of %s.\n", 
		dictname, err, cause);
}

static int get_nentry_parent(struct dictnode *p)
{
	p = p->dn_parent;
	while(p->dn_type != DNTYPE_LIST)
		p = p->dn_parent;
	return p->dn_number;
}

void
dictlink_test(booklist books)
{
	char buf[1024];	
	int i;
	strlist_iterator itr;
	struct dictnode *cur = NULL, *next;
	struct dlistent *dep;

	dictlink_init(books, test_handler);

	printf("cur: null\n");
	while(fgets(buf, 1024, stdin) != NULL){
		buf[strlen(buf)-1] = '\0';
		if(buf[0] == 'h'){
			printf("u: 	up node\n");
			printf("n: 	next entry\n");
			printf("p: 	previous entry\n");
			printf("lword: 	lookup word and set current\n");
			printf("eentryid: go translation entry\n");
			printf("tN: 	set current node to Nth entry in current word entry\n");
			printf("s: 	show contents of current entry\n");
			printf("d: 	dump\n");
			printf("q: 	quit\n");
		}

		else if(buf[0] == 'u'){
			if(!cur || !(next = dl_up(books, cur)))
				printf("can't up\n");
			else 
				cur = next;
		}
		else if(buf[0] == 'p'){
			if(!cur || !(next = dl_prev(books, cur)))
				printf("can't prev\n");
			else 
				cur = next;
		}
		else if(buf[0] == 'n'){
			if(!cur || !(next = dl_next(books, cur)))
				printf("can't next\n");
			else 
				cur = next;
		}

		else if(buf[0] == 'l'){
			if(buf[1] == '\0')
				printf("lword\n");
			else {
				next = dl_lookup(books, &buf[1], cur);
				if(next == NULL)
					printf("no such word: %s\n", &buf[1]);
				else 
					cur = next;
			}
		}

		else if(buf[0] == 'e'){
			if(!cur)
				printf("cur is not set\n");
			else if(buf[1] == '\0')
				printf("eentryid\n");

			else {
				next = dl_entry(books, &buf[1], cur);
				if(next == NULL)
					printf("no such word: %s\n", &buf[1]);
				else 
					cur = next;
			}
		}
		else if(buf[0] == 't'){
			if(!cur)
				printf("cur is not set\n");
			else if(buf[1] == '\0')
				printf("tN\n");
			else if(cur->dn_type != DNTYPE_LIST)
				printf("%s is not list.\n", cur->dn_key);
			else {
				dep = dl_list_at(cur, atoi(&buf[1]));
				if(dep){
					next = dl_entry(books, 
							dep->dl_key,
							cur);
					if(next)
						cur = next;
				}

			}
		}

		else if(buf[0] == 's'){
			if(!cur)
				printf("cur is not set\n");

			else if(cur->dn_type == DNTYPE_LIST){
				for(i = 0; i < dl_list_count(cur); ++i){
					dep = dl_list_at(cur, i);
					printf("%d: %s<%s>\t%s\n", i, 
						dep->dl_title, dep->dl_key, 
						dep->dl_book->db_glid);
				}
			}
			else {
				printf("  %s %d/%d: %s\n", cur->dn_whist->whe_word, 
					cur->dn_number + 1, 
					get_nentry_parent(cur),
					cur->dn_book->db_glid);

				itr = iterator_first(cur->dn_content);
				for(; !iterator_is_end(itr); itr = iterator_next(itr))
					printf("- %s\n", iterator_data(itr));
			}
		}
		else if(buf[0] == 'd')
			dictnode_dump(cur, 0);

		else if(buf[0] == 'D'){
			void tcp_set_error(int);
			if(buf[1] == '\0')
				printf("DN\n");
			else 
				tcp_set_error(atoi(&buf[1]));
		}
		else if(buf[0] == 'q')
			break;

		printf("cur: %s   ", (cur ? cur->dn_key: "null"));
		if(cur){
			if(dl_has_up(cur))
				printf("u");
			if(dl_has_prev(cur))
				printf("p");
			if(dl_has_next(cur))
				printf("n");
		}
		printf("\n");
	}

	dl_quit(books, cur);
	dictlink_fin(books);
}
