/*
 *			main.c
 * $Id: main.c,v 1.12 2004/01/19 22:16:19 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifndef XPOPDIC_VERSION
#define XPOPDIC_VERSION		"test"
#endif /* XPOPDIC_VERSION */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "debug.h"
#include "memory.h"
#include "tcp.h"
#include "bitmap.h"
#include "property.h"
#include "dictbook.h"
#include "ndtpbook.h"
#include "ndtpserver.h"
#include "subr.h"
#include "xwindow.h"
#include "stk/stkbase.h"
#include "xpopdic.h"

#include "ndtpsrvlist.c"

int xp_debug_flags = 0;

static int  booklistanalysis(ndtpsrvlist, booklist);
static int isnumberstr(const char *p);
static struct ndtpserver *getserver(ndtpsrvlist, const char *);
static int getbook(struct ndtpserver *server, const char *, booklist);

/* test mode */
void dictlink_test(booklist);

int
main(int argc, char **argv)
{
	ndtpsrvlist  		srvlist;
	booklist		books;
	booklist_iterator	bitr;
	const char 		*dispname = NULL;
	struct property		*prp = NULL;
	xrootwindow		*xrwp = NULL; 

	/* initialize property database */
	property_init();	

	/* parse and analysis options */
	optanalysis(argc, argv);

	/* is debug mode */
	if(property_get("Debug") != NULL)
		DEBUG_ENABLE(DEBUG_ALL);

	if(DEBUG_IS_ENABLED(DEBUG_ALL))
		property_dump();

	/* open root window */
	dispname = getenv("DISPLAY");

	/* analysis booklist and open books */
	srvlist = ndtpsrvlist_new(4, 4, xpdic_malloc, xpdic_free);
	books =   booklist_new(4, 4, xpdic_malloc, xpdic_free);
	
	/* is there one book at least? */
	if(!booklistanalysis(srvlist, books) || iterator_count(books) == 0){
		fprintf(stderr, "No available book!\n");
		aggregator_delete(books);
		aggregator_delete_by(srvlist, ndtpserver_free);
		property_fin();
		memdump();
		exit(1);
	}

	/* show book list */
	if(DEBUG_IS_ENABLED(DEBUG_ALL)){
		struct dictbook *bp; 
		struct ndtpbook *nbp; 
		bitr = iterator_first(books);
		for(; !iterator_is_end(bitr); bitr = iterator_next(bitr)){
			bp = iterator_data(bitr);
			nbp = (struct ndtpbook *)iterator_data(bitr);

			printf("%d %s\n", nbp->ndb_index, bp->db_glid);
		}
	}

	/* always dictionary open? */
	prp = property_get("AlwaysOpen");
	if(prp && prty_integer_val(prp) == 0){
		/* this option is disable */
		ndtpsrvlist_iterator	sitr;
		sitr = iterator_first(srvlist);

		/* close all dictionaries. */
		for(; !iterator_is_end(sitr); sitr = iterator_next(sitr))
			ndtpserver_disconnect(iterator_data(sitr));
	}

	/* is test mode */
	if(property_get("Test") != NULL)
		dictlink_test(books);

	else { /* start xpopdic */
		if(!(xrwp = xrootwindow_open(dispname))){
			fprintf(stderr, "can't open display: %s\n", dispname);
			exit(1);
		}
		xpopdic_start(xrwp, books);
		xrootwindow_close(xrwp);
	}

	/* free books */
	aggregator_delete_by(books, dict_free);

	/* free NDTP servers */
	aggregator_delete_by(srvlist, ndtpserver_free);
	property_fin();
	memdump();
	return 0;
}


static int
booklistanalysis(ndtpsrvlist srvs, booklist books)
{
	char  			bookname[1024], srvname[256], *bp, *sp, *ep;
	struct property 	*prop;

	/* get dictionary list from property db */
	prop = property_get("Dictionary");
	strncpy(bookname, prty_string_val(prop), 1024);
	bp = bookname;

	do {
		struct ndtpserver *server = NULL;
		sp = index(bp, '@');
		if(!sp){
			fprintf(stderr, "bad format: %s\n", 
				 prty_string_val(prop));
			return 0;
		}
		*sp++ = '\0';
		ep = index(sp, ',');

		if(ep)
			*ep++ = '\0';

		/* copy server name into buffer */
		strncpy(srvname, sp, 256);
		if(!index(sp, ':'))
			strcat(srvname, ":2010");
		if(!(server = getserver(srvs, srvname))){
			fprintf(stderr, "try next....\n");
			continue;
		}

		/* book name or index indication */
		if(getbook(server, bp, books) == 0){
			fprintf(stderr, "NDTP server %s doesn't have book \"%s\".\n",
				 server->ns_id, bp);
			fprintf(stderr, "try next....\n");
			continue;
		}
	} while((bp = ep) != '\0');
	return 1;
}


static int isnumberstr(const char *p)
{
	while(*p)
		if(*p > '9' || *p < '0')
			return 0;
		else
			++p;

	return 1;
}


static struct ndtpserver *
getserver(ndtpsrvlist srvs, const char *srvname)
{
	int ret;
	ndtpsrvlist_iterator itr;
	struct ndtpserver *server;

	/* search serverlist */
	itr = iterator_first(srvs);
	for(; !iterator_is_end(itr); itr = iterator_next(itr))
		if(!strcmp(srvname, iterator_data(itr)->ns_id))
			return iterator_data(itr);

	if((server = ndtpserver_alloc(srvname)) == NULL){
		fprintf(stderr, "ndtpserver_alloc(%s): failed.\n", srvname);
		return NULL;
	}

	if((ret = ndtpserver_connect(server)) <= 0){
		fprintf(stderr, "ndtpserver_connect(%s): failed(%d):", srvname, ret);
		if(ret < 0)
			perror(" ");
		else
			printf("\n");
		ndtpserver_free(server);
		return NULL;
	}

	if((ret = ndtpserver_load(server)) <= 0){
		fprintf(stderr, "ndtpserver_load(%s): failed(%d):", srvname, ret);
		if(ret < 0)
			perror(" ");
		else
			printf("\n");
		ndtpserver_free(server);
		return NULL;
	}
	iterator_append(srvs, server);
	return server;
}

static int
getbook(struct ndtpserver *server, const char *bookname, booklist books)
{
	int idx;
	booklist_iterator itr;
	struct ndtpbook *nbp;

	if(isnumberstr(bookname)){
		idx = atoi(bookname) - 1;

		if(iterator_count(server->ns_books) <= idx)
			return NULL;
		itr = iterator_at(server->ns_books, idx);
		iterator_append(books, iterator_data(itr));
		return 1;
	}

	/* put all books into the book list */
	if(!strcasecmp("all", bookname)){
		itr = iterator_first(server->ns_books);
		for(; !iterator_is_end(itr); itr = iterator_next(itr))
			iterator_append(books, iterator_data(itr));
		return iterator_count(books);
	}

	itr = iterator_first(server->ns_books);
	for(; !iterator_is_end(itr); itr = iterator_next(itr)){
		nbp = (struct ndtpbook *)iterator_data(itr);
		if(!strcasecmp(nbp->ndb_id, bookname)){
			iterator_append(books, iterator_data(itr));
			return 1;
		}
	}
	return 0;
}

