/*
 *			tcp.c
 * $Id: tcp.c,v 1.8 2004/01/12 06:57:59 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include "tcp.h"

extern int errno;

void 
tc_disconnect(struct tcpcon *tc)
{
	close(tc->tc_soc);
	tc->tc_soc = -1;
	tc->tc_ptr = NULL;
	tc->tc_datalen = 0;
}

static int 
tc_fillbuf(struct tcpcon *tc)
{
	/* read from a connection */
	int nread; 
again:
	nread = read(tc->tc_soc, tc->tc_buf, TC_BUFSIZ);
	if(nread < 0 && errno == EINTR)
		goto again;
	else if(nread < 0)
		return -1;
	tc->tc_datalen = nread;
	tc->tc_ptr = tc->tc_buf;
	return nread;
}

static int
tc_readbuf(struct tcpcon *tc, char *bufp, int len)
{
	int		ret = 0;

	/* read from buffered data */
	while(tc->tc_datalen > 0 && len > 0){
		int c = *tc->tc_ptr;
		*bufp++ = *tc->tc_ptr++;
		--tc->tc_datalen;  /* decrement buffer for the connection */
		--len;		   /* decrement length of user buffer */
		++ret;		   /* nread */

		if(c == '\n'){
			*bufp = '\0';
			return ret;
		}
	}
	return ret;
}

#ifdef DEBUG_DICTLINK
static int err = 0;
void tcp_set_error(int i){ err = i; printf("err -> %d\n", i);}
#define CHKERR(x)	do { 			\
		if(err != 0){			\
			(x)->tc_soc = -1;	\
			errno = err;		\
			return -1;		\
		}				\
	} while(0)
#else 
void tcp_set_error(int i){  }
#define CHKERR(x)	
#endif 

int  
tc_readLine(struct tcpcon *tc, char *origbuf, int origlen)
{
	int		nread = 0, total = 0;
	char 		*bufp = origbuf;
	int		len = origlen;

	CHKERR(tc);
	
	while(1){
		/* check buffer first */
		if(tc->tc_datalen == 0){
			if((nread = tc_fillbuf(tc)) < 0)
				return -1;	/* error */
			else if(nread == 0)
				return total;	/* closed by peer */
		}

		nread = tc_readbuf(tc, bufp, len);
		total += nread;

		/* user buffer gets full */
		if(len == nread)
			return total;

		/* nread must be less than len */
		if(bufp[nread - 1] == '\n')
			return total;	/* have read successfully */

		if(tc->tc_datalen != 0)
			abort();	/* something wrong */

		len -= nread;
		bufp += nread;
	}

	/* never reach */
	return -1;
}

int 
tc_writeLine(struct tcpcon *tc, const char *buf, int len)
{
	CHKERR(tc);
	return write(tc->tc_soc, buf, len);
}


#ifndef HAVE_GETADDRINFO
// IPv4 only 
#include <netinet/in.h>
int  
tc_connect(struct tcpcon *tc, const char *dest, short port)
{
	struct hostent *hp;
	struct sockaddr_in sin;

	errno = 0;
	memset(tc, 0, sizeof(struct tcpcon));

	hp = gethostbyname(dest);
	if(!hp)
		return 0;

	if((tc->tc_soc = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return 0;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);

	if(connect(tc->tc_soc, (struct sockaddr*)&sin, sizeof(sin)) < 0){
		close(tc->tc_soc);
		tc->tc_soc = -1;
		return 0;
	}
	return 1;
}

int
tc_is_connected(struct tcpcon *tc)
{
	struct sockaddr_in peer;
	int len = sizeof(peer);

	if(tc->tc_soc < 0)
		return 0;

	if(getpeername(tc->tc_soc, (struct sockaddr *)&peer, &len) < 0)
		return 0;
	return 1;
}

#else // HAVE_GETADDRINFO
int  
tc_connect(struct tcpcon *tc, const char *dest, short port)
{
	int soc = -1, n;
	struct addrinfo hints, *res, *ressave;
	char serv[128];

	errno = 0;

	memset(tc, 0, sizeof(struct tcpcon));
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	snprintf(serv, 127, "%u", port);

	if((n = getaddrinfo(dest, serv, &hints, &res)) != 0)
		return 0;

	ressave = res;

	do {
		soc = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
		if(soc < 0)
			continue;

		if(connect(soc, res->ai_addr, res->ai_addrlen) == 0)
			break;

		close(soc);
	} while((res = res->ai_next) != NULL);

	if(res == NULL)
		return 0;
	freeaddrinfo(ressave);

	tc->tc_soc = soc;
	return 1;
}

int
tc_is_connected(struct tcpcon *tc)
{
	struct sockaddr_storage peer;
	int len = sizeof(peer);

	if(tc->tc_soc < 0)
		return 0;
	if(getpeername(tc->tc_soc, (struct sockaddr *)&peer, &len) < 0)
		return 0;
	return 1;
}
#endif /* _XP_TCP_H_ */

#ifdef TCP_TEST
int
main(void){
	int nread;
	struct tcpcon tc;
	char buf[1024];

	if(!tc_connect(&tc, "192.168.0.2", 80)){
		if(errno == 0)
			fprintf(stderr, "tc_connect(): can't resolve\n");
		else
			perror("tc_connect");
	}

	printf("is con? = %d\n", tc_is_connected(&tc));
	if(tc_writeLine(&tc, "GET/\n", 5) != 5){
		perror("tc_writeLine");
		exit(1);
	}
	printf("is con? = %d\n", tc_is_connected(&tc));

	if((nread = tc_readLine(&tc, buf, 1024)) < 0){
		perror("tc_readLine");
		exit(1);
	}
	write(1, buf, nread);

	close(tc.tc_soc);

	printf("is con? = %d\n", tc_is_connected(&tc));
	return 0;
}
#endif
