/*
 *			bitmap.c
 * $Id: bitmap.c,v 1.1 2004/01/14 15:48:02 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bitmap.h"
#include "memory.h"

	/* bitmap utilities */
static char hexchr[256];
static int init_bitmap_done = 0;

static void
init_bitmap(void)
{
	register int i;
	memset(hexchr, -1, sizeof(hexchr));

	for(i = '0' ; i <= '9' ; ++i)
		hexchr[i] = i - '0';
	for(i = 'a' ; i <= 'f' ; ++i)
		hexchr[i] = i - 'a' + 10;
	for(i = 'A' ; i <= 'F' ; ++i)
		hexchr[i] = i - 'A' + 10;
}

static int
get_last_value(const char *buf)
{
	char *p = strrchr(buf, ' ');
	if(!p)
		p = strrchr(buf, '\t');
	if(!p)
		return -1;
	return atoi(p+1);
}

static void
parse_bitmap_line(char *bp, int *posp, const char *linep)
{
	unsigned char val;
	const char *p = linep;

	/* initialize hexadeciaml character table */
	if(!init_bitmap_done){
		init_bitmap();
		init_bitmap_done = 1;
	}

	while((p = strstr(p, "0x"))){
		val = hexchr[(int)*(p+2)] << 4 | hexchr[(int)*(p+3)]; 
		p += 4;
		bp[*posp] = val;
		++(*posp);
	}
}


/* parse_bitmap() parses bitmap files and allocates memory for the bitmap */
struct bitmap *
bitmap_parse(char *(*nextline)(void *, char *, int), void *arg)
{
	int  x, y, nbytes;
	int hotx = -1, hoty = -1;
	int  pos = 0;
	char buf[1024];
	char *bmp;
	struct bitmap *ret;

	if(!nextline(arg, buf, 1024) ||
	   (x = get_last_value(buf)) == -1)
		return NULL;	/* bad format */

	if(!nextline(arg, buf, 1024) ||
	   (y = get_last_value(buf)) == -1)
		return NULL;	/* bad format */

	nbytes = x * y / 8;
	bmp = MALLOC("bitmap/data", nbytes);

	/* first line must contain "static unsigned char ... \n" or 
		"#define default_x_hot ..." */
	if(!nextline(arg, buf, 1024))
		goto bad;

	/* hotspot is defined? */
	if(!strncmp("#define", buf, 7)){
		/* this file defines hotstop */
		hotx = get_last_value(buf);
		if(!nextline(arg, buf, 1024))
			goto bad;
		hoty = get_last_value(buf);
	}
	
	while(pos != nbytes){
		if(!nextline(arg, buf, 1024))
			goto bad;
		parse_bitmap_line(bmp, &pos, buf);
	}

	ret = MALLOC("bitmap", sizeof(struct bitmap));
	ret->bm_data = bmp;
	ret->bm_width = x;
	ret->bm_height = y;
	ret->bm_hotx = hotx;
	ret->bm_hoty = hoty;
	return ret;
bad:
	FREE("bitmap/data", bmp);
	return NULL;
}

static char *
nextline_file(void *vp, char *buf, int size) 
{ 
	return fgets(buf, size, (FILE *)vp); 
}

struct bitmap *
bitmap_load(FILE *fp){ return bitmap_parse(nextline_file, fp); }

struct bitmap *
bitmap_scale(struct bitmap *srcdata, int dstw, int dsth)
{
	register int i, j;	
	int x, y, srcw, srch;
	int newsize = (dstw + 7) / 8 * dsth;
	struct bitmap *ret;
	unsigned char *dstp, *srcp;

	/* initialize */
	/* dstp = MALLOC("bitmap/data", (dstw * dsth / 8)); */
	dstp = MALLOC("bitmap/data", newsize);

	srcw = srcdata->bm_width;
	srch = srcdata->bm_height;
	srcp = srcdata->bm_data;

	for(i = 0; i < newsize ; ++i)
		dstp[i] = 0;

	for(i = 0; i < dsth; ++i)
		for(j = 0; j < dstw ; ++j){
			int  srcidx, dstidx, srcbits, dstbits;
			char srcmask = 0, dstmask = 0;

			x=(int)(((double)srcw/(double)dstw)*j);
			y=(int)(((double)srch/(double)dsth)*i);

			//y * srcw + x 
			srcbits = (srcw + 7) / 8;
			srcidx = (y * srcbits) + (x / 8);
			srcmask = 1 << x % 8;

			dstbits = (dstw + 7) / 8;
			dstidx = (i * dstbits) + (j / 8);
			dstmask = 1 <<  j % 8;

			if(srcp[srcidx] & srcmask)
				dstp[dstidx] |= dstmask;

		}

	ret = MALLOC("bitmap", sizeof(struct bitmap));
	ret->bm_data = dstp;
	ret->bm_width = dstw;
	ret->bm_height = dsth;

	if(srcdata->bm_hotx != -1 && srcdata->bm_hoty){
		ret->bm_hotx = (int)(((double)srcw/(double)dstw)*srcdata->bm_hotx);
		ret->bm_hoty = (int)(((double)srcw/(double)dstw)*srcdata->bm_hoty);
	}
	else {
		ret->bm_hotx = -1;
		ret->bm_hoty = -1;
	}
	return ret;
}

void
bitmap_free(struct bitmap *bmp)
{
	FREE("bitmap/data", bmp->bm_data);
	FREE("bitmap", bmp);
}

int
bitmap_store(struct bitmap *bmp, FILE *fp)
{
	register int i;
	int size = (bmp->bm_width + 7) / 8 * bmp->bm_height;
	fprintf(fp, "#define\tdefault_width\t%d\n", bmp->bm_width);
	fprintf(fp, "#define\tdefault_height\t%d\n", bmp->bm_height);

	if(bmp->bm_hotx != -1 && bmp->bm_hoty != -1){
		fprintf(fp, "#define\tdefault_x_hot\t%d\n", bmp->bm_hotx);
		fprintf(fp, "#define\tdefault_y_hot\t%d\n", bmp->bm_hoty);
	}

	printf("static unsigned char default_bits[] = {\n");
	for(i = 0; i < size ; i++){
		fprintf(fp, "0x%02x", bmp->bm_data[i] & 0xff);
		if(i != size - 1)
			fprintf(fp, ",%c", ((i + 1) % 8) == 0? '\n': ' ');
	}
	fprintf(fp, "};\n");
	return 1;
}


#ifdef BITMAP_TEST 
int
main(int argc, char **argv)
{
	int i = 0;
	int x, y, x1, y1;
	FILE *fp;
	struct bitmap *bmp1, *bmp2;

	if(argc != 3){
		fprintf(stderr, "Usage: %s size filename\n", *argv);
		return 1;
	}

	if(sscanf(argv[1], "%dx%d", &x, &y) != 2){
		fprintf(stderr, "size format(8x16) error: %s\n", argv[1]);
		return 1;
	}

	if((fp = fopen(argv[2], "r")) == NULL){
		perror(argv[2]);
		return 1;
	}

	if((bmp1 = bitmap_load(fp)) == 0){
		fprintf(stderr, "load_bitmap error: malformed data\n");
		return 1;
	}
	fclose(fp);

	bmp2 = bitmap_scale(bmp1, x, y);
	bitmap_store(bmp2, stdout);

	bitmap_free(bmp1);
	bitmap_free(bmp2);

	memdump();

	return 0;
}

#endif /* BITMAP_TEST */
