/*
 *			xwindow.c
 * $Id: xwindow.c,v 1.7 2004/02/11 22:56:13 hiddy Exp $
 * Copyright (C) 2003 Hideaki Imaizumi <hiddy@sfc.wide.ad.jp>
 *               All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "memory.h"
#include "bitmap.h"
#include "xwindow.h"

/* static variable */
#include "pointer.bmp"
#include "pointermask.bmp"
static struct bitmap pointer_bmp = { pointer_width, pointer_height, 
				     pointer_x_hot, pointer_y_hot,
				     pointer_bits };
static struct bitmap pointermask_bmp = { pointermask_width, pointermask_height, 
				     pointermask_x_hot, pointermask_y_hot,
				     pointermask_bits };

/*-------------------* X Root Window *--------------------*/
struct _xrootwindow {
	struct xrootwindow _base;
	Display		   *xr_disp;
	xcursor 	   *xr_cursor;
};
#define XRWDisplay(x) ((struct _xrootwindow *)x)->xr_disp
#define XRWRootWin(x)  RootWindow(XRWDisplay(x), 0)

xrootwindow *
xrootwindow_open(const char *dispname)
{
	struct _xrootwindow *ret;
	XWindowAttributes       rootattr;
	Display 		*d;

	/* open display */
	if(!(d = XOpenDisplay(dispname)))
		return NULL;

	/* clear cut buffer. */
	XStoreBuffer(d, " ", 2, 0);

	/* allocate memory for root window */
	ret = MALLOC("xrootwindow", sizeof(struct _xrootwindow));
	ret->xr_disp = d;

	ret->xr_cursor = xcursor_new((xrootwindow *)ret, &pointer_bmp, 
			  &pointermask_bmp, pointer_x_hot, pointer_y_hot, 
			  0, 0xffffff);

	/* get width/height of root window */
	XGetWindowAttributes(d, RootWindow(d, 0), &rootattr);
	ret->_base.xr_width = rootattr.width;
	ret->_base.xr_height = rootattr.height;

	return (xrootwindow *)ret;
}

void
xrootwindow_close(xrootwindow *xrp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)xrp;
	xcursor_free(xrp, xrwp->xr_cursor);
	XCloseDisplay(xrwp->xr_disp);
	FREE("xrootwindow", xrwp);
}

void
xroot_flush(xrootwindow *_xrp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xrp;
	XFlush(xrwp->xr_disp);
}

void
xroot_getptr(xrootwindow *_xrp, int *x, int *y)
{
	Window		root, child;
	int		wx, wy;
	unsigned int	mask;
	struct _xrootwindow *xrmp = (struct _xrootwindow *)_xrp;

	XQueryPointer(xrmp->xr_disp, RootWindow(xrmp->xr_disp, 0),
		      &root, &child, x, y, &wx, &wy, &mask);
}

const char *
xroot_cutbuffer(xrootwindow *_xrp, int *nbytes)
{
	struct _xrootwindow *xrmp = (struct _xrootwindow *)_xrp;
	char *p = XFetchBytes(xrmp->xr_disp, nbytes);

	if(p)
		p[*nbytes] = '\0';
	return p;
}

void 
xroot_clear_cutbuffer(xrootwindow *_xrp)
{
	struct _xrootwindow *xrmp = (struct _xrootwindow *)_xrp;
	XStoreBuffer(xrmp->xr_disp, " ", 2, 0);
}
/*---------------------------------------------------------*/

/*-------------------* X Window Attr *--------------------*/
struct _xattr {  
	xrootwindow 	*xatr_root;
	GC 		xatr_gc; 
	XFontStruct	*xatr_fontinfo;
};
#define XATRGC(x)  	((struct _xattr *)x)->xatr_gc
#define XATRFontInfo(x) ((struct _xattr *)x)->xatr_fontinfo
#define XATRRoot(x)  	((struct _xattr *)x)->xatr_root
#define XATRDisplay(x) 	XRWDisplay(XATRRoot(x))
#define XATRWindow(x)  	XRWRootWin(XATRRoot(x))
/*---------------------------------------------------------*/

/*--------------------* X Image  *--------------------*/
struct _ximage { 
	xrootwindow 	*xi_root;
	/* XImage 		*xi_image;  */
	Pixmap 		xi_pixmap;
};
/*---------------------------------------------------------*/

/*--------------------* X Image  *--------------------*/
struct _xcursor { 
	Cursor		xc_cursor;
};
/*---------------------------------------------------------*/

/*----------------------* X Window *-----------------------*/
struct _xwindow {
	struct xwindow		_base;
	union {
		struct 	{
			Window	_win;	  /* front window */
			Window	_shadow;  /* shadow window */
		} _w;
		struct {
			Pixmap	_pmap;	  /* pixmap */
			xattr  *_atr;	  /* default attribute */
		} _p;
	} _ud;
	int			xw_xdiff, xw_ydiff;/* offset for shadow */
	XSetWindowAttributes 	xw_wattr;  /* win atrribute */
	xcursor			*xw_cursor;	/* current cursor */
	long			xw_bgc;	   /* background color */
	long			xw_bdc;	   /* border color */
	Window			xw_focus;	/* previously focused window */
};
#define XWIsWindow(x) (((struct xwindow *)x)->xw_type == XWTYPE_WINDOW)
#define XWWindow(x) (XWIsWindow(x)? ((struct _xwindow *)x)->xw_win: \
				    ((struct _xwindow *)x)->xw_pixmap)
#define XWDisplay(x) XRWDisplay((x)->xw_root)

#define xw_win		_ud._w._win
#define xw_shadow	_ud._w._shadow
#define xw_pixmap	_ud._p._pmap
#define xw_pattr	_ud._p._atr

/* control function declarations for X window */
xwindow *
xwindow_create(xrootwindow *_xrp, int x, int y, 
		unsigned int w, unsigned int h, 
		int bdwidth, long bgcolor, long bdcolor, 
		int shadow_enable, int winmgr_disable)
{
	static const long shadowcol = 0x003f3f3f; /* shadow color */
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xrp;
	Display *d = xrwp->xr_disp;
	struct _xwindow *xwp;
	int	rever;

	/* allocate memory for xwindow */
	xwp = MALLOC("xwindow", sizeof(struct _xwindow));

	/* set some attributes of window */
	xwp->_base.xw_type = XWTYPE_WINDOW;
	xwp->_base.xw_root = _xrp;
	xwp->_base.xw_x = x;
	xwp->_base.xw_y = y;
	xwp->_base.xw_width = w;
	xwp->_base.xw_height = h;
	xwp->xw_xdiff = ((w / 25) < 3? 3: w/25);
	xwp->xw_ydiff = ((h / 25) < 3? 3: h/25);
	xwp->xw_bgc = bgcolor;
	xwp->xw_bdc = bdcolor;

	xwp->xw_win = XCreateSimpleWindow(d, RootWindow(d,0), x, y, w, h, 
					     bdwidth, bdcolor, bgcolor);

	xwp->xw_shadow = NULL;
	if(shadow_enable)
		xwp->xw_shadow = XCreateSimpleWindow(d, RootWindow(d,0), 
					x + xwp->xw_xdiff, y + xwp->xw_ydiff, 
					w, h, bdwidth, shadowcol, shadowcol);

	if(winmgr_disable){
		/* Disable Window Manager. */
		xwp->xw_wattr.override_redirect = True;
		if(xwp->xw_shadow)
			XChangeWindowAttributes(d,xwp->xw_shadow, CWOverrideRedirect,
						&xwp->xw_wattr);
		XChangeWindowAttributes(d,xwp->xw_win, CWOverrideRedirect,
					&xwp->xw_wattr);
	}

	xw_raise((xwindow *)xwp);

	xw_set_cursor((xwindow *)xwp, xrwp->xr_cursor);
	XGetInputFocus(d, &xwp->xw_focus, &rever); 
	return (xwindow *)xwp;
}


xwindow *
xpixmap_create(xwindow *pp,  long bgcolor, unsigned int w, unsigned int h)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)pp->xw_root;
	Display *d = xrwp->xr_disp;
	struct _xwindow *parentp = (struct _xwindow *)pp;
	struct _xwindow *xwp;

	/* allocate memory for xwindow */
	xwp = MALLOC("xwindow", sizeof(struct _xwindow));

	/* set some attributes of window */
	xwp->_base.xw_type = XWTYPE_PIXMAP;
	xwp->_base.xw_root = (struct xrootwindow *)xrwp;
	xwp->_base.xw_x = 0;
	xwp->_base.xw_y = 0;
	xwp->_base.xw_width = w;
	xwp->_base.xw_height = h;
	xwp->xw_bgc = bgcolor;
	xwp->xw_xdiff = xwp->xw_ydiff = 0;
	xwp->xw_pattr = xattr_alloc((struct xrootwindow *)xrwp);
	xattr_set_fgcolor(xwp->xw_pattr, bgcolor);
	xattr_set_bgcolor(xwp->xw_pattr, bgcolor);

	xwp->xw_pixmap = XCreatePixmap(d, XWWindow(parentp), w, h, DefaultDepth(d,0));
	xw_clear((xwindow *)xwp);
	return (xwindow *)xwp;
}

void 
xwindow_destroy(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;

	if(_xwp->xw_type == XWTYPE_WINDOW){
		xw_unmap(_xwp);

       		XDestroyWindow(xrwp->xr_disp, xwp->xw_win);
	        if(xwp->xw_shadow)
       	 		XDestroyWindow(xrwp->xr_disp, xwp->xw_shadow);
	 }
	 else if(_xwp->xw_type == XWTYPE_PIXMAP){
		xattr_free(xwp->xw_pattr);
		XFreePixmap(xrwp->xr_disp, xwp->xw_pixmap);
	}

	FREE("xwindow", _xwp);
}

void xw_map(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;

        if(xwp->xw_shadow)
        	XMapWindow(xrwp->xr_disp, xwp->xw_shadow);
        XMapWindow(xrwp->xr_disp, xwp->xw_win);
}

void xw_unmap(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;

        if(xwp->xw_shadow)
        	XUnmapWindow(xrwp->xr_disp, xwp->xw_shadow);
        XUnmapWindow(xrwp->xr_disp, xwp->xw_win);
}

void xw_focus(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	int 	rever;

	XGetInputFocus(xrwp->xr_disp, &xwp->xw_focus, &rever); 
	XSetInputFocus(xrwp->xr_disp, xwp->xw_win, RevertToNone, CurrentTime);
}

void xw_focus_return(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	XSetInputFocus(xrwp->xr_disp, xwp->xw_focus, RevertToNone, CurrentTime);
}

void xw_raise(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	Window wins[2];

        XRaiseWindow(xrwp->xr_disp, xwp->xw_win);
        if(xwp->xw_shadow){
		wins[0] = xwp->xw_win;
		wins[1] = xwp->xw_shadow;
		XRestackWindows(xrwp->xr_disp, wins, 2);
        	/*XRaiseWindow(xrwp->xr_disp, xwp->xw_shadow); */
	}
}

void xw_clear(xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;

	if(_xwp->xw_type == XWTYPE_WINDOW)
	        XClearWindow(xrwp->xr_disp, xwp->xw_win);
	else 
		xw_clear_region(_xwp, 0, 0, _xwp->xw_width, _xwp->xw_height);
}

static int next_x(int mx, xwindow *_xwp);
static int next_y(int my, xwindow *_xwp);

void xw_move(xwindow *_xwp, int x, int y)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	int nextx = next_x(x, _xwp);
	int nexty = next_y(y, _xwp);

	/* move shadow window */
	if(xwp->xw_shadow)
		XMoveResizeWindow(xrwp->xr_disp, xwp->xw_shadow, 
				  nextx + xwp->xw_xdiff, nexty + xwp->xw_ydiff, 
				  _xwp->xw_width, _xwp->xw_height);
	XMoveResizeWindow(xrwp->xr_disp, xwp->xw_win, 
				  nextx, nexty, _xwp->xw_width, _xwp->xw_height);
}

long xw_get_bgcolor(xwindow *_xwp)
{
	return ((struct _xwindow *)_xwp)->xw_bgc;
}
long xw_get_bdcolor(xwindow *_xwp)
{
	return ((struct _xwindow *)_xwp)->xw_bgc;
}

void xw_set_cursor(xwindow *_xwp, xcursor *_xcur)
{
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xcursor *xcp = (struct _xcursor *)_xcur;

	if(!_xcur)
		xw_set_cursor(_xwp, xrwp->xr_cursor);
	else {
		XDefineCursor(xrwp->xr_disp, xwp->xw_win, xcp->xc_cursor);	
		xwp->xw_cursor = _xcur;
	}
}

xcursor *xw_get_cursor(xwindow *_xwp)
{
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	return xwp->xw_cursor;
}

void xw_discard_events(xwindow *_xwp)
{
	XEvent e;
	while(XCheckMaskEvent(XWDisplay(_xwp), 
		PointerMotionMask | ButtonMotionMask, &e))
		XNextEvent(XWDisplay(_xwp), &e);
}


void xw_clear_region(xwindow *_xwp, int x, int y, unsigned int w, unsigned int h)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;

	if(_xwp->xw_type == XWTYPE_WINDOW)
	        XClearArea(xrwp->xr_disp, xwp->xw_win, x, y, w, h, False);
	else 
		xw_fill_rectangle(_xwp, xwp->xw_pattr, x, y, w, h);
}

void xw_copy_region(xwindow *_src, xwindow *_dst, int srcx, int srcy, 
		  unsigned int w, unsigned int h, int dstx, int dsty)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_src->xw_root;
	struct _xwindow *src = (struct _xwindow *)_src;
	struct _xwindow *dst = (struct _xwindow *)_dst;

	if(_src->xw_type == XWTYPE_PIXMAP)
		XCopyArea(xrwp->xr_disp, src->xw_win, dst->xw_win, 
			  XATRGC(src->xw_pattr), srcx, srcy, w, h, dstx, dsty);
	else 
		XCopyArea(xrwp->xr_disp, src->xw_win, dst->xw_win, 
			  DefaultGC(xrwp->xr_disp, DefaultScreen(xrwp->xr_disp)),
			  srcx, srcy, w, h, dstx, dsty);
}

/*---------------------------------------------------------*/


/*---------------* X Window Draw Functions *---------------*/
/* draw function declarations for X window */
void xw_draw_point(xwindow *w, struct xattr *a, int x, int y)
{
	XDrawPoint(XWDisplay(w), XWWindow(w), XATRGC(a), x, y);
}
		
void xw_draw_line(xwindow *w, struct xattr *a, 
	     int x1, int y1, int x2, int y2)
{
	XDrawLine(XWDisplay(w), XWWindow(w), XATRGC(a), x1, y1, x2, y2);
}

void xw_draw_rectangle(xwindow *wp, struct xattr *a, int x, int y, 
		       unsigned int w, unsigned int h)
{
	XDrawRectangle(XWDisplay(wp), XWWindow(wp), XATRGC(a), x, y, w, h);
}

void xw_draw_arc(xwindow *wp, struct xattr *a, int x, int y, 
	   	 unsigned int w, unsigned int h, int a1, int a2)
{
	XDrawArc(XWDisplay(wp), XWWindow(wp), XATRGC(a), x, y, w, h, a1, a2);
}

void xw_draw_image(xwindow *xw, struct xattr *a, int x, int y,
		   ximage *imgp, int w, int h)
{
	struct _ximage *ximp = (struct _ximage *)imgp;

	/*
	XPutImage(XWDisplay(xw), XWWindow(xw), XATRGC(a), (XImage *)ximp->xi_image,
		  0, 0,  x, y, w, h);
	*/
	XCopyPlane(XWDisplay(xw), ximp->xi_pixmap, XWWindow(xw), XATRGC(a), 0, 0, 
		  w, h, x, y, 1);
}

void xw_draw_string(xwindow *w, struct xattr *a, int x, int y,
		    const char *str, int size)
{
	XDrawString(XWDisplay(w), XWWindow(w), XATRGC(a), x, 
		    y + XATRFontInfo(a)->max_bounds.ascent, str, size);
}

static void euc2jis(char ec1, char ec2, char* jc1, char* jc2);
void xw_draw_wstring(xwindow *w, struct xattr *a, int x, int y,
		     const char *cp, int size)
{
	register  int 	i;
	char		tmp[1500], *bufp = tmp;

	if((size * 2) > (int)sizeof(tmp))
		bufp = MALLOC("dstrbuf", size * 2 + 8);

        for(i = 0; i < size; i++)
                euc2jis(cp[i * 2], cp[(i * 2) + 1],
                        &bufp[i * 2], &bufp[(i * 2) + 1]);
	XDrawString16(XWDisplay(w), XWWindow(w), XATRGC(a), 
		      x, y + XATRFontInfo(a)->max_bounds.ascent, 
		      (XChar2b*)bufp, size);

        if(bufp != tmp)
		FREE("dstrbuf", bufp);
}
		/* these functions draws background indicated by xattr */
void xw_draw_string_bg(xwindow *w, struct xattr *a, int x, int y,
		       const char *str, int size)
{
	XDrawImageString(XWDisplay(w), XWWindow(w), XATRGC(a), x, 
			 y + XATRFontInfo(a)->max_bounds.ascent, str, size);
}

void xw_draw_wstring_bg(xwindow *w, struct xattr *a, int x, int y,
			const char *cp, int size)
{
	register  int 	i;
	char		tmp[1500], *bufp = tmp;

	if((size * 2) > (int)sizeof(tmp))
		bufp = MALLOC("dstrbuf", size * 2 + 8);

        for(i = 0; i < size; i++)
                euc2jis(cp[i * 2], cp[(i * 2) + 1],
                        &bufp[i * 2], &bufp[(i * 2) + 1]);
	XDrawImageString16(XWDisplay(w), XWWindow(w), XATRGC(a), 
		      x, y + XATRFontInfo(a)->max_bounds.ascent, 
		     (XChar2b*)bufp, size);

        if(bufp != tmp)
		FREE("dstrbuf", bufp);
}

void xw_fill_rectangle(xwindow *wp, struct xattr *a, int x, int y, 
		       unsigned int w, unsigned int h)
{
	XFillRectangle(XWDisplay(wp), XWWindow(wp), XATRGC(a), x, y, w, h);
}
void xw_fill_arc(xwindow *wp, struct xattr *a, int x, int y, 
		 unsigned int w, unsigned int h, int a1, int a2)
{
	XFillArc(XWDisplay(wp), XWWindow(wp), XATRGC(a), x, y, w, h, a1, a2);
}


/*
void xw_fill_polygon(xwindow *, struct xattr *, xpoint *, 
				int npoints, enum plyg_shape, enum plyg_mode);
*/

static int 
next_x(int mx, xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	int w = _xwp->xw_width;

	if(mx < 0)
		return 0;

	/* shadow window is enabled */
	if(xwp->xw_shadow != NULL)
		w += xwp->xw_xdiff;

	if(mx + w > xrwp->_base.xr_width)
		return xrwp->_base.xr_width - w;

	return mx;
}

static int next_y(int my, xwindow *_xwp)
{
	struct _xrootwindow *xrwp = (struct _xrootwindow *)_xwp->xw_root;
	struct _xwindow *xwp = (struct _xwindow *)_xwp;
	int h = _xwp->xw_height;

	if(my < 0)
		return 0;

	/* shadow window is enabled */
	if(xwp->xw_shadow != NULL)
		h += xwp->xw_ydiff;

	if(my + h > xrwp->_base.xr_height)
		return xrwp->_base.xr_height - h;

	return my;
}


/* misc */
static void euc2jis(char ec1, char ec2, char* jc1, char* jc2)
{
	static unsigned int z[64] = {
	0x2121,0x2123,0x2156,0x2157,0x2122,0x2126,0x2572,0x2521,
	0x2523,0x2525,0x2527,0x2529,0x2563,0x2565,0x2567,0x2543,
	0x213c,0x2522,0x2524,0x2526,0x2528,0x252a,0x252b,0x252d,
	0x252f,0x2531,0x2533,0x2535,0x2537,0x2539,0x253b,0x253d,
	0x253f,0x2541,0x2544,0x2546,0x2548,0x254a,0x254b,0x254c,
	0x254d,0x254e,0x254f,0x2552,0x2555,0x2558,0x255b,0x255e,
	0x255f,0x2560,0x2561,0x2562,0x2564,0x2566,0x2568,0x2569,
	0x256a,0x256b,0x256c,0x256d,0x256f,0x2573,0x212b,0x212c };

	// this is hankaku.
	if(((ec1 & 0xff) == 0x8e)){
		if(ec2 < (char)0xa0 && ec2 > (char)0xdf){
			// strange character;
			*jc1 = *jc2 = 0;
			return;
		}
		*jc1 = (z[ec2 - 0xa0] >> 8) & 0xff;
		*jc2 = z[ec2 - 0xa0]  & 0xff;
	}
	else {
		*jc1 = (ec1 & ~0x80);
		*jc2 = (ec2 & ~0x80);
	}
}
/*---------------------------------------------------------*/


/*-----------------* X Window Attributes *-----------------*/
xattr *
xattr_alloc(xrootwindow *_xrp)
{
	struct _xattr *ret = MALLOC("xattr", sizeof(struct _xattr));
	ret->xatr_gc = XCreateGC(XRWDisplay(_xrp), 
				 XRWRootWin(_xrp), 0, 0);
	ret->xatr_fontinfo = NULL;
	ret->xatr_root = _xrp;
	return (xattr *)ret;
}

void
xattr_free(struct xattr *_xatp)
{
	struct _xattr *xatp = (struct _xattr *)_xatp;

	if(xatp->xatr_fontinfo){
		XFreeFont(XATRDisplay(xatp), xatp->xatr_fontinfo);
		xatp->xatr_fontinfo = NULL;
	}
	XFreeGC(XATRDisplay(xatp), xatp->xatr_gc);
	FREE("xattr", _xatp);
}

/* 
 * functions to set attributes into xattr structure 
 *  - return value: succeed = 1, fail = 0 
 */
int xattr_set_font(xattr *a, const char *c)
{
	Font font = XLoadFont(XATRDisplay(a), c);
	if(!font)
		return 0;

	if(XATRFontInfo(a))
		XFreeFont(XATRDisplay(a), XATRFontInfo(a));
	XATRFontInfo(a) = XQueryFont(XATRDisplay(a), font);
	XSetFont(XATRDisplay(a), XATRGC(a), font);
	return 1;
}

int xattr_get_font_width(xattr *a)
{
	if(!XATRFontInfo(a))
		return 0;
	return XATRFontInfo(a)->max_bounds.width;
}

int xattr_get_font_height(xattr *a)
{
	if(!XATRFontInfo(a))
		return 0;
	return XATRFontInfo(a)->max_bounds.ascent + 	
		XATRFontInfo(a)->max_bounds.descent;
}

int
xattr_get_str_width(xattr *a, const char *s, int n)
{
	int direction, ascent, descent;
	XCharStruct xcs;

	if(!XATRFontInfo(a))
		return 0;

	XQueryTextExtents(XATRDisplay(a), XATRFontInfo(a)->fid, 
			  s, n, &direction, &ascent, &descent, &xcs);

	/*
	printf("a=%d, d=%d, xsc.a=%d, xsc.d=%d, h=%d\n", 
		ascent, descent, xcs.ascent, xcs.descent, 
			xattr_get_font_height(a));
	*/
	return xcs.width;
}
int
xattr_get_wstr_width(xattr *a, const char *cp, int size)
{
	register  int 	i;
	char		tmp[1500], *bufp = tmp;
	int 		direction, ascent, descent;
	XCharStruct 	xcs;

	if((size * 2) > (int)sizeof(tmp))
		bufp = MALLOC("dstrbuf", size * 2 + 8);

        for(i = 0; i < size; i++)
                euc2jis(cp[i * 2], cp[(i * 2) + 1],
                        &bufp[i * 2], &bufp[(i * 2) + 1]);

	XQueryTextExtents16(XATRDisplay(a), XATRFontInfo(a)->fid, 
			  (XChar2b *)bufp, size, &direction, &ascent, 
			  &descent, &xcs);
        if(bufp != tmp)
		FREE("dstrbuf", bufp);
	return xcs.width;
}

int xattr_get_fgcolor(xattr *a, long *color)
{
	XGCValues   xgv;
	if(!color)
		return 1;
	XGetGCValues(XATRDisplay(a), XATRGC(a), GCForeground, &xgv);
	*color = xgv.foreground;
	return 1;
}

int xattr_get_bgcolor(xattr *a, long *color)
{
	XGCValues   xgv;
	if(!color)
		return 1;
	XGetGCValues(XATRDisplay(a), XATRGC(a), GCBackground, &xgv);
	*color = xgv.background;
	return 1;
}

int xattr_set_fgcolor(xattr *a, long color)
{
	XGCValues   xgv;
	xgv.foreground = color;
	XChangeGC(XATRDisplay(a), XATRGC(a), GCForeground, &xgv);
	return 1;
}

int xattr_set_bgcolor(xattr *a, long color)
{
	XGCValues   xgv;
	xgv.background = color;
	XChangeGC(XATRDisplay(a), XATRGC(a), GCBackground, &xgv);
	return 1;
}

int xattr_set_fgcolor_byname(xattr *a, const char *name)
{
	Display *d = XATRDisplay(a);
	XColor xc, xco;
	XAllocNamedColor(d, DefaultColormap(d, 0), name, &xc, &xco);
	return xattr_set_fgcolor(a, xc.pixel);
}

int xattr_set_bgcolor_byname(xattr *a, const char *name)
{
	Display *d = XATRDisplay(a);
	XColor xc, xco;
	XAllocNamedColor(d, DefaultColormap(d, 0), name, &xc, &xco);
	return xattr_set_bgcolor(a, xc.pixel);
}

int xattr_set_line_width(xattr *a, int w)
{
	XGCValues   xgv;
	xgv.line_width = w;
	XChangeGC(XATRDisplay(a), XATRGC(a), GCLineWidth, &xgv);
	return 1;
}

int xattr_set_line_style(xattr *a, enum xattr_line_style s)
{
	XGCValues   xgv;
	xgv.line_style = 0;

	switch(s){
	case XatLineSolid:
		xgv.line_style = LineSolid;
		break;
	case XatLineDoubleDash:
		xgv.line_style = LineDoubleDash;
		break;
	case XatLineOnOffDash:
		xgv.line_style = LineOnOffDash;
		break;
	}
	XChangeGC(XATRDisplay(a), XATRGC(a), GCLineStyle, &xgv);
	return 1;
}

int xattr_set_cap_style(xattr *a, enum xattr_cap_style s)
{
	XGCValues   xgv;
	xgv.cap_style = 0;

	switch(s){
	case XatCapNotLast:
		xgv.cap_style = CapNotLast;
		break;
	case XatCapButt:
		xgv.cap_style = CapButt;
		break;
	case XatCapRound:
		xgv.cap_style = CapRound;
		break;
	case XatCapProjecting: 
		xgv.cap_style = CapProjecting;
		break;
	}

	XChangeGC(XATRDisplay(a), XATRGC(a), GCCapStyle, &xgv);
	return 1;
}

int xattr_set_join_style(xattr *a, enum xattr_join_style s)
{
	XGCValues   xgv;
	xgv.join_style = 0;
	switch(s){
	case XatJoinMiter:
		xgv.join_style = JoinMiter;
		break;
	case XatJoinRound:
		xgv.join_style = JoinRound;
		break;
	case XatJoinBevel:
		xgv.join_style = JoinBevel;
		break;
		
	}
	XChangeGC(XATRDisplay(a), XATRGC(a), GCJoinStyle, &xgv);
	return 1;
}


int xattr_set_fill_style(xattr *a, enum xattr_fill_style s)
{
	XGCValues   xgv;
	xgv.fill_style = 0;

	switch(s){
	case XatFillSolid:
		xgv.fill_style = FillSolid;
		break;
	case XatFillTiled:
		xgv.fill_style = FillTiled;
		break;
	case XatFillOpaqueStippled:
		xgv.fill_style = FillOpaqueStippled;
		break;
	case XatFillStippled:
		xgv.fill_style = FillStippled;
		break;
			
	}
	XChangeGC(XATRDisplay(a), XATRGC(a), GCFillStyle, &xgv);
	return 1;
}

int xattr_set_fill_rule(xattr *a, enum xattr_fill_rule s)
{
	XGCValues   xgv;
	xgv.fill_rule = 0;
	switch(s){
	case XatFillEvenOddRule: 
		xgv.fill_rule = EvenOddRule;
		break;
	case XatFillWindingRule:
		xgv.fill_rule = WindingRule;
		break;
	}

	XChangeGC(XATRDisplay(a), XATRGC(a), GCFillRule, &xgv);
	return 1;
}

int xattr_set_arc_mode(xattr *a, enum xattr_arc_mode s)
{
	XGCValues   xgv;
	xgv.arc_mode = 0;
	switch(s){
	case XatArcPieSlice:
		xgv.arc_mode = ArcPieSlice;
		break;
	case XatArcChord:
		xgv.arc_mode = ArcChord;
		break;
	}

	XChangeGC(XATRDisplay(a), XATRGC(a), GCArcMode, &xgv);
	return 1;
}

/*--------------------* X Image  *--------------------*/
ximage *
ximage_from_bitmap(xrootwindow *rp, struct bitmap *bmp)
{
	struct _xrootwindow *xrp = (struct _xrootwindow *)rp;
	Display *d = xrp->xr_disp;
	struct _ximage *ret = MALLOC("ximage", sizeof(struct _ximage));
	/*
	Visual *visual = DefaultVisual(xrp->xr_disp, 0);
	ret->xi_image = XCreateImage(d, visual, 1, XYBitmap, 0, bmp->bm_data, 
				bmp->bm_width, bmp->bm_height, 8, (bmp->bm_width + 7)/ 8);
	*/
	ret->xi_root = rp;
	ret->xi_pixmap = XCreatePixmapFromBitmapData(d,RootWindow(d, 0), 
				bmp->bm_data, bmp->bm_width, bmp->bm_height,
				WhitePixel(d,0),BlackPixel(d,0),DefaultDepth(d,0)); 
	return (ximage *)ret;
}
void 
ximage_free(ximage *imgp)
{
	struct _ximage *ximp = (struct _ximage *)imgp;
	struct _xrootwindow *xrp = (struct _xrootwindow *)ximp->xi_root;
	/* 
		calling XDestroyImage() causes 
	"xpopdic in free(): warning: chunk is already free."
		XDestroyImage(ximp->image); 
	*/
	XFreePixmap(xrp->xr_disp, ximp->xi_pixmap);
	FREE("ximage", imgp);
}

/*--------------------* X Cursor  *--------------------*/
xcursor *
xcursor_new(xrootwindow *_xrw, struct bitmap *bmp,
	    struct bitmap *mask, int x_hot, int y_hot,
            long fgcolor, long bgcolor)
{
	struct _xrootwindow *xrw = (struct _xrootwindow *)_xrw;
	Display *d = xrw->xr_disp;
	int xhot, yhot;
	XColor fg, bg, fgo, bgo;
	char fgname[8], bgname[8];
	struct _xcursor *ret = NULL;
	Pixmap c, m;

	xhot = (x_hot == -1)? 8 : x_hot;
	yhot = (y_hot == -1)? 8 : y_hot;

	/* size check */
	if(!bmp || !mask 
	 || bmp->bm_width != mask->bm_width 
	 || bmp->bm_height != mask->bm_height)
		return NULL; 
	
	if(xhot < 0 || yhot < 0 
	|| xhot >= bmp->bm_width || yhot >= bmp->bm_height)
		return NULL;

	snprintf(fgname, 8, "#%06lx", fgcolor);
	snprintf(bgname, 8, "#%06lx", bgcolor);
	XAllocNamedColor(d, DefaultColormap(d, 0), fgname, &fg, &fgo);
	XAllocNamedColor(d, DefaultColormap(d, 0), bgname, &bg, &bgo);

	c = XCreatePixmapFromBitmapData(d,RootWindow(d, 0), 
			bmp->bm_data, bmp->bm_width, bmp->bm_height,
			BlackPixel(d, 0), WhitePixel(d, 0), 1);
	m = XCreatePixmapFromBitmapData(d,RootWindow(d, 0), 
			mask->bm_data, mask->bm_width, mask->bm_height,
			WhitePixel(d, 0), BlackPixel(d, 0), 1);

	ret = MALLOC("xcursor", sizeof(struct _xcursor));
	ret->xc_cursor = XCreatePixmapCursor(d, c, m, &fg, &bg, xhot, yhot);
	XFreePixmap(d, c);
	XFreePixmap(d, m);
	return (xcursor *)ret;
}

void
xcursor_free(xrootwindow *_xrw, xcursor *_xcur)
{
	struct _xrootwindow *xrw = (struct _xrootwindow *)_xrw;
	Display *d = xrw->xr_disp;
	struct _xcursor *xcur = (struct _xcursor *)_xcur;

	XFreeCursor(d, xcur->xc_cursor);
	FREE("xcursor", _xcur);
}



/*-----------------* X Window Event *-----------------*/

static unsigned long mkevmask(struct xw_event_handler *evp)
{
	unsigned long evmask = 0;
	if(evp->xeh_button_pressed)
		evmask |= ButtonPressMask;
	if(evp->xeh_button_released)
		evmask |= ButtonReleaseMask;
	if(evp->xeh_button_motion)
		evmask |= PointerMotionMask | ButtonMotionMask;
	if(evp->xeh_key_pressed)
		evmask |= KeyPressMask;
	if(evp->xeh_key_released)
		evmask |= KeyReleaseMask;
	if(evp->xeh_expose)
		evmask |= ExposureMask;
	if(evp->xeh_visibility_changed)
		evmask |= VisibilityChangeMask;
	if(evp->xeh_focus_in && evp->xeh_focus_out)
		evmask |= FocusChangeMask;
	if(evp->xeh_mouse_in)
		evmask |= EnterWindowMask;
	if(evp->xeh_mouse_out)
		evmask |= LeaveWindowMask;

	return evmask;
}

static int keystate2modifier(int s)
{
	int ret = 0;
	if(s & ShiftMask)
		ret |= XKM_SHIFT;
	if(s & LockMask)
		ret |= XKM_CAPS;
	if(s & ControlMask)
		ret |= XKM_CTRL;
	if(s & Mod1Mask)
		ret |= XKM_META;
	if(s & Button1Mask)
		ret |= XKM_BUTTON1;
	if(s & Button2Mask)
		ret |= XKM_BUTTON2;
	if(s & Button3Mask)
		ret |= XKM_BUTTON3;
	return ret;
}

static int call_evhandler(xwindow *_xwp, struct xw_event_handler *evp,
			  XEvent *xevp)
{
	int button = 0;
	KeySym keysym;
	char key;

	switch(xevp->type){
	case ButtonPress:
		button = (xevp->xbutton.button == Button1? 0:
			   (xevp->xbutton.button == Button2? 1: 2));
		return evp->xeh_button_pressed(_xwp, evp, 
			xevp->xbutton.x, xevp->xbutton.y, button);
	case ButtonRelease:
		button = (xevp->xbutton.button == Button1? 0:
			   (xevp->xbutton.button == Button2? 1: 2));
		return evp->xeh_button_released(_xwp, evp, 
			xevp->xbutton.x, xevp->xbutton.y, button);
	case MotionNotify:
		button = -1;
		if(xevp->xmotion.state & Button1Mask)
			button = 0;
		else if(xevp->xmotion.state & Button2Mask)
			button = 1;
		else if(xevp->xmotion.state & Button3Mask)
			button = 2;
		return evp->xeh_button_motion(_xwp, evp, 
			xevp->xmotion.x, xevp->xmotion.y, button);
		
	case KeyPress:
		XLookupString((XKeyEvent *)xevp, &key, 1, &keysym, NULL);
		return evp->xeh_key_pressed(_xwp, evp, 
			xevp->xkey.x, xevp->xkey.y, XKeysymToString(keysym),
			keystate2modifier(xevp->xkey.state));
	case KeyRelease:
		XLookupString((XKeyEvent *)xevp, &key, 1, &keysym, NULL);
		return evp->xeh_key_released(_xwp, evp, 
			xevp->xkey.x, xevp->xkey.y, XKeysymToString(keysym),
			keystate2modifier(xevp->xkey.state));
	case Expose:
		if(evp->xeh_expose)
			return evp->xeh_expose(_xwp, evp);
		break;
	case NoExpose:
		if(evp->xeh_noexpose)
			return evp->xeh_noexpose(_xwp, evp);
		return 1; /* ignore */

	case VisibilityNotify:
		if(evp->xeh_visibility_changed)
			return evp->xeh_visibility_changed(_xwp, evp, 
				xevp->xvisibility.state);
		return 1; /* ignore */
	case EnterNotify:
		if(evp->xeh_mouse_in)
			return evp->xeh_mouse_in(_xwp, evp, xevp->xcrossing.x, 
						 xevp->xcrossing.y);
		return 1; /* ignore */
	case LeaveNotify:
		if(evp->xeh_mouse_out)
			return evp->xeh_mouse_out(_xwp, evp, xevp->xcrossing.x, 
						  xevp->xcrossing.y);
		return 1; /* ignore */
	case FocusIn:
		if(evp->xeh_focus_in)
			return evp->xeh_focus_in(_xwp, evp);
		return 1; /* ignore */
	case FocusOut:
		if(evp->xeh_focus_out)
			return evp->xeh_focus_out(_xwp, evp);
		return 1; /* ignore */
	case SelectionClear:
	case SelectionRequest:
	case SelectionNotify:
		return 1;	/* ignore */
	}

	printf("XEvent: Non-handling Event(%d) is received.\n", xevp->type); 
	return 1; /* unknown event */
}

int
xw_has_event(xwindow *_xwp, struct xw_event_handler *evp)
{
	unsigned long evmask =  mkevmask(evp);
	XEvent xev;

	XSelectInput(XWDisplay(_xwp), XWWindow(_xwp), evmask);
	return XCheckMaskEvent(XWDisplay(_xwp), evmask, &xev);
}

int
xw_process_event(xwindow *_xwp, struct xw_event_handler *evp)
{
	unsigned long evmask =  mkevmask(evp);
	XEvent xev;

	/* 
	 *  XSelectInput must be done by calling xw_has_event() before
	 * calling this function 
	 */
	while(XCheckMaskEvent(XWDisplay(_xwp), evmask, &xev)){
		XNextEvent(XWDisplay(_xwp), &xev);
		if(!call_evhandler(_xwp, evp, &xev))
			return 0;
	} 
	return 1;
}

void 
xw_event_loop(xwindow *_xwp, struct xw_event_handler *evp, 
	      void (*before)(void), void (*after)(void))
{
	unsigned long evmask = mkevmask(evp);
	XEvent xev;

	XSelectInput(XWDisplay(_xwp), XWWindow(_xwp), evmask);

	while(1){
		int ret = 0;
		XNextEvent(XWDisplay(_xwp), &xev);

		if(before)	before(); 
		/* critical region */
		ret = call_evhandler(_xwp, evp, &xev);
		/* /critical region */
		if(after)	after();

		if(!ret)
			break;
	}
}

void            
xw_stop_event_loop(xwindow *xwp)
{
#if 0
	XButtonEvent ev;

	memset(&ev, 0, sizeof(ev));
	ev.type = ButtonPress;
	ev.display = XWDisplay(xwp);
	ev.window = XWWindow(xwp);
	ev.root = RootWindow(ev.display, 0);
	ev.button = Button2;
	ev.time = CurrentTime;
	ev.x = 1;
	ev.y = 1;
	ev.x_root = 1;
	ev.y_root = 1;
	ev.same_screen = True;
	XSendEvent(XWDisplay(xwp), XWDisplay(xwp), True, 
		   ButtonPressMask, (XEvent *)&ev);
#endif
}


/* X selection */
struct _xselection {
	struct xselection _base;
	Display		  *d;
	Window		   w;
	Atom		   sel;
	Atom		   target;
};

struct xselection *
xselection_new(struct xrootwindow *r, int sel)
{
	struct _xselection *xsp;
	Atom	sels[2] = { XA_PRIMARY, XA_SECONDARY };
	Display *d = ((struct _xrootwindow *)r)->xr_disp;
	Window   w;
	XSetWindowAttributes attr;

	if(sel != XSEL_PRIMARY && sel != XSEL_SECONDARY)	
		return NULL;

	w = XCreateSimpleWindow(d, RootWindow(d, 0), 0, 0, 1, 1,
				0, 0, 0);
	attr.override_redirect = True;
	XChangeWindowAttributes(d, w, CWOverrideRedirect, &attr);

	xsp = MALLOC("xselection", sizeof(struct _xselection));

	xsp->_base.xsl_sel = sel;
	xsp->d = d;
	xsp->w = w;
	xsp->sel = sels[sel];
	xsp->target = XInternAtom(d, "__COPY_TEXT", False);

	return (struct xselection *)xsp;
}

void
xselection_free(struct xselection *_xsp)
{
	struct _xselection *xsp = (struct _xselection *)_xsp;
	XDeleteProperty(xsp->d, xsp->w, xsp->target);
	XDestroyWindow(xsp->d, xsp->w);	
	FREE("xselection", xsp);
}

static int
xsel_receive(Display *d, Window w, Atom sel, Atom target, char *b, int len)
{
	time_t t;
	int ntry = 10;
	XEvent e;
	Atom type;      
	unsigned char *buf;  
	int  fmt;
	unsigned long nitems = 0, left;

	time(&t);

	/* request property */
	XConvertSelection(d, sel, XA_STRING, 
			  target, w, t);

	/* avoid block */
	while(XCheckTypedEvent(d, SelectionNotify, &e) != True 
		&& ntry-- > 0)
		;

	/* give up. nobody owns selectoin PRIMARY */
	if(ntry == 0)
		return 0;

	if(e.xselection.selection != sel ||
	   e.xselection.property != target)
		return 0;

	if(XGetWindowProperty(d, w, target, 0, 1024, 
	   False, XA_STRING, &type, &fmt, &nitems, &left, &buf)
		== Success && type == XA_STRING && nitems > 0){
		strncpy(b, buf, (len < nitems ? len: nitems));

		b[nitems < len? nitems: len - 1] = '\0';
		XFree(buf); 
		return ((nitems < len) ? nitems : len);
	}
	return 0;
}

int
xselection_get(struct xselection *_xsp, char *b, int len)
{
	struct _xselection *xsp = (struct _xselection *)_xsp;
	return xsel_receive(xsp->d, xsp->w, xsp->sel, xsp->target, b, len);
}


#ifdef XW_TEST
int
main(void)
{
	xrootwindow *xroot;
	xwindow *a, *b;
	xattr *at;

	xroot = xrootwindow_open(NULL);
	at = xattr_alloc(xroot);
	xattr_set_fgcolor(at, 0xf0);
	xattr_set_bgcolor(at, 0xffffff);
	xattr_set_font(at, "fixed");

	b = xwindow_create(xroot, 100, 100, 200, 100, 3, 0xffffff, 0xffffff, 1, 1);
	a = xpixmap_create(b, 0xffffff, 400, 200);
	xw_map(b);
	xroot_flush(xroot);
	getchar();

	xw_draw_string(a, at, 10, 10, "YogaHead!!!!", 12);
	xw_fill_rectangle(a, at, 20, 100, 150, 80);
	xroot_flush(xroot);

	xw_copy_region(a, b, 0, 0, 200, 100, 0, 0);
	xroot_flush(xroot);

	getchar();
	xw_copy_region(a, b, 0, 100, 200, 100, 0, 0);
	xroot_flush(xroot);

	getchar();
	xw_copy_region(a, b, 0, 0, 200, 100, 0, 0);
	xroot_flush(xroot);

	getchar();
	xw_clear(a);
	getchar();
	xwindow_destroy(a);
	xwindow_destroy(b);
	xattr_free(at);
	xrootwindow_close(xroot);
	stk_memdump();
	return 0;
}
#endif
